#include <string.h>
#include <math.h>
#include <lal/LALStdlib.h>
#include <lal/Comm.h>
#include <lal/RingSearch.h>
#include <BuildDB.h>
#include <WrapperRingSearch.h>

#define NTMPLT 10 /* number of templates to filter per slave call */

NRCSID( APPLYSEARCHC, PKG "$Id: ApplySearch.c,v 1.2 2002/06/06 19:13:56 jolien Exp $" );

enum { ExchRingTmplts, ExchNumRingTmpltsDone, ExchFinished };

void
BuildOutput(
    LALStatus       *status,
    LALSearchOutput *output,
    RingEventList   *eventList
    );

void
LALApplySearch(
    LALStatus             *status,
    LALSearchOutput       *output,
    LALSearchInput        *input,
    LALApplySearchParams  *applyParams
    )
{
  InitExchParams    initExchParams;
  RingSearchParams *params;

  INITSTATUS( status, "LALApplySearch", APPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

  ASSERT( output, status,
      WRAPPERRINGSEARCHH_ENULL, WRAPPERRINGSEARCHH_MSGENULL );
  ASSERT( input, status,
      WRAPPERRINGSEARCHH_ENULL, WRAPPERRINGSEARCHH_MSGENULL );
  ASSERT( applyParams, status,
      WRAPPERRINGSEARCHH_ENULL, WRAPPERRINGSEARCHH_MSGENULL );
  ASSERT( applyParams->searchParams, status,
      WRAPPERRINGSEARCHH_ENULL, WRAPPERRINGSEARCHH_MSGENULL );

  params = applyParams->searchParams;

  if ( params->myProcNumber < 0 )
  {
    int rank;
    MPI_Comm_rank( *applyParams->mpiParams->comm, &rank );
    params->myProcNumber = rank;
  }

  initExchParams.mpiComm   = *applyParams->mpiParams->comm;
  initExchParams.myProcNum = params->myProcNumber;

  if ( params->searchMaster ) /* search master */
  {
    INT4 rtrn = 0;

    if ( params->numSlaves < 0 )
    {
      int size;
      MPI_Comm_size( initExchParams.mpiComm, &size );
      params->numSlaves = size - 1;
    }

    while ( params->numSlaves > 0 && ! rtrn )
    {
      ExchParams *thisExch = NULL;

      LALInitializeExchange( status->statusPtr, &thisExch, NULL,
          &initExchParams );
      CHECKSTATUSPTR( status );

      switch ( thisExch->exchObjectType )
      {
        case ExchRingTmplts:
          {
            UINT4 todo = NTMPLT;
            if ( todo > params->templateBank->numTmplt - params->templatesSent )
              todo = params->templateBank->numTmplt - params->templatesSent;
            LALExchangeUINT4( status->statusPtr, &params->templatesSent,
                thisExch );
            CHECKSTATUSPTR( status );
            LALExchangeUINT4( status->statusPtr, &todo, thisExch );
            CHECKSTATUSPTR( status );
            params->templatesSent += todo;
          }
          break;
        case ExchNumRingTmpltsDone:
          {
            UINT4 done;
            LALExchangeUINT4( status->statusPtr, &done, thisExch );
            CHECKSTATUSPTR( status );
            params->templatesDone += done;
          }
          rtrn = 1; /* break out of loop */
          break;
        case ExchFinished:
          {
            UINT4 numEvents;
            LALExchangeUINT4( status->statusPtr, &numEvents, thisExch );
            CHECKSTATUSPTR( status );
            params->numEvents += numEvents;
          }
          --params->numSlaves;
          break;
        default:
          ABORT( status, WRAPPERRINGSEARCHH_EEXCH,
              WRAPPERRINGSEARCHH_MSGEEXCH );
      }
      LALFinalizeExchange( status->statusPtr, &thisExch );
      CHECKSTATUSPTR( status );
    }
    
    if ( params->numSlaves )
    {
      output->fracRemaining  = params->templateBank->numTmplt;
      output->fracRemaining -= params->templatesDone;
      output->fracRemaining /= params->templateBank->numTmplt;
      output->notFinished    = 1;
      output->numOutput      = 0;
      output->result         = NULL;
    }
    else /* master populates search_summary table */
    {
      SearchSummaryInfo info;
      UINT4             seq;
      INT8              ns;
      int               size;
      MPI_Comm_size( *applyParams->mpiParams->comm, &size );
      info.dsoName = "ldasring";
      /* get start/end time from "channel" sequence */
      for ( seq = 0; seq < input->numberSequences; ++seq )
      {
        multiDimData *sequence = input->sequences + seq;
        if ( ! strstr( sequence->name, "channel" ) )
          continue;
        info.inputStartTime.gpsSeconds     = sequence->range.dTime.startSec;
        info.inputStartTime.gpsNanoSeconds = sequence->range.dTime.startNan;
        info.inputEndTime.gpsSeconds       = sequence->range.dTime.stopSec;
        info.inputEndTime.gpsNanoSeconds   = sequence->range.dTime.stopNan;
        break;
      }
      /* output times differ from input times by a quarter of a segment */
      ns  = (INT8)1000000000 * (INT8)info.inputStartTime.gpsSeconds;
      ns += (INT8)info.inputStartTime.gpsNanoSeconds;
      ns += (INT8)floor( 1e9 * (REAL8)params->segmentSize
          / ( 4.0 * params->sampleRate ) );
      info.outputStartTime.gpsSeconds     = ns / (INT8)1000000000;
      info.outputStartTime.gpsNanoSeconds = ns % (INT8)1000000000;
      ns  = (INT8)1000000000 * (INT8)info.inputEndTime.gpsSeconds;
      ns += (INT8)info.inputEndTime.gpsNanoSeconds;
      ns -= (INT8)floor( 1e9 * (REAL8)params->segmentSize
          / (4.0 * params->sampleRate ) );
      info.outputEndTime.gpsSeconds     = ns / (INT8)1000000000;
      info.outputEndTime.gpsNanoSeconds = ns % (INT8)1000000000;
      info.numEvents = params->numEvents;
      info.numNodes  = size;
      output->fracRemaining       = 0;
      output->notFinished         = 0;
      output->numOutput           = 1;
      output->result              = LALCalloc( 1, sizeof( *output->result ) );
      output->result->search      = ringDown;
      output->result->significant = 1;
      LALCreateSearchSummaryDataBase( status->statusPtr,
          &output->result->results, &info );
      CHECKSTATUSPTR( status );
    }
  }
  else /* search slave */
  {
    UINT4 startTemplate  = 0;
    UINT4 templatesToDo  = 0;
    ExchParams     exchTmplts;
    ExchParams     exchNumDone;
    ExchParams     exchFinished;
    ExchParams    *thisExch  = NULL;
    RingEventList *eventList = NULL;

    exchTmplts.exchObjectType = ExchRingTmplts;
    exchTmplts.send           = 0; /* master sends */
    exchTmplts.numObjects     = 2; /* two numbers  */
    exchTmplts.partnerProcNum = 0; /* master */

    exchNumDone.exchObjectType = ExchNumRingTmpltsDone;
    exchNumDone.send           = 1; /* I send */
    exchNumDone.numObjects     = 1; /* one number */
    exchNumDone.partnerProcNum = 0; /* master */

    exchFinished.exchObjectType = ExchFinished;
    exchFinished.send           = 1; /* I send */
    exchFinished.numObjects     = 1; /* one number */
    exchFinished.partnerProcNum = 0; /* master */

    /* get the templates to filter */
    LALInitializeExchange( status->statusPtr, &thisExch, &exchTmplts,
        &initExchParams );
    CHECKSTATUSPTR( status );
    LALExchangeUINT4( status->statusPtr, &startTemplate, thisExch );
    CHECKSTATUSPTR( status );
    LALExchangeUINT4( status->statusPtr, &templatesToDo, thisExch );
    CHECKSTATUSPTR( status );
    LALFinalizeExchange( status->statusPtr, &thisExch );
    CHECKSTATUSPTR( status );

    if ( ! templatesToDo ) /* nothing left to do: tell master shutting down */
    {
      LALInitializeExchange( status->statusPtr, &thisExch, &exchFinished,
          &initExchParams );
      CHECKSTATUSPTR( status );
      LALExchangeUINT4( status->statusPtr, &params->numEvents, thisExch );
      CHECKSTATUSPTR( status );
      LALFinalizeExchange( status->statusPtr, &thisExch );
      CHECKSTATUSPTR( status );
      output->notFinished = 0; /* finished */
    }
    else /* OK, do the filtering! */
    {
      RingSearchInput searchin;
      searchin.startTemplate = startTemplate;
      searchin.templatesToDo = templatesToDo;
      LALRingSearch( status->statusPtr, &eventList, &searchin, params );
      CHECKSTATUSPTR( status );
      LALInitializeExchange( status->statusPtr, &thisExch, &exchNumDone,
          &initExchParams );
      CHECKSTATUSPTR( status );
      LALExchangeUINT4( status->statusPtr, &templatesToDo, thisExch );
      CHECKSTATUSPTR( status );
      LALFinalizeExchange( status->statusPtr, &thisExch );
      CHECKSTATUSPTR( status );
      output->notFinished = 1; /* not finished */
    }

    output->fracRemaining = 0; /* ignored */

    if ( eventList )
    {
      BuildOutput( status->statusPtr, output, eventList );
      CHECKSTATUSPTR( status );
      while ( eventList ) /* free memory */
      {
        RingEventList *next = eventList->next;
        LALFree( eventList );
        eventList = next;
      }
    }
  }

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


void
BuildOutput(
    LALStatus       *status,
    LALSearchOutput *output,
    RingEventList   *eventList
    )
{
  const size_t ifoNameLen = 2;
  RingEventList *thisEvent;
  UINT4 event;
  UINT4 numEvents = 0;
  outPut   *theOutput;
  dataBase *dbEntry;

  INITSTATUS( status, "BuildOutput", APPLYSEARCHC );

  thisEvent = eventList;
  while ( thisEvent )
  {
    ++numEvents;
    thisEvent = thisEvent->next;
  }

  output->numOutput = 1;
  theOutput = output->result = LALCalloc( 1, sizeof( *output->result ) );
  if ( ! output->result )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  
  theOutput->search      = ringDown;
  theOutput->significant = 1;
  theOutput->results     = LALCalloc( 1, sizeof( *theOutput->results ) );
  if ( ! theOutput->results )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }

  /*
   *
   * IFO names.
   *
   */

  dbEntry = theOutput->results;
  strncpy( dbEntry->tableName, "sngl_ringdown", sizeof( dbEntry->tableName ) );
  strncpy( dbEntry->columnName, "ifo", sizeof( dbEntry->columnName ) );
  dbEntry->type       = char_s_ptr;
  dbEntry->numberRows = numEvents;
  dbEntry->rowDimensions = LALCalloc( dbEntry->numberRows,
      sizeof( *dbEntry->rowDimensions ) );
  dbEntry->rows.chars    = LALCalloc( dbEntry->numberRows * ifoNameLen,
      sizeof( *dbEntry->rows.chars ) );
  thisEvent = eventList;
  for ( event = 0; event < numEvents; ++event )
  {
    dbEntry->rowDimensions[event] = ifoNameLen;
    memcpy( dbEntry->rows.chars + ifoNameLen * event, thisEvent->ifoName,
        ifoNameLen * sizeof( *dbEntry->rows.chars ) );
    thisEvent = thisEvent->next;
  }


  /*
   *
   * Start time seconds.
   *
   */

  dbEntry->next           = LALCalloc( 1, sizeof( *dbEntry->next ) );
  if ( ! dbEntry->next )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  dbEntry->next->previous = dbEntry;
  dbEntry                 = dbEntry->next;

  strncpy( dbEntry->tableName, "sngl_ringdown", sizeof( dbEntry->tableName ) );
  strncpy( dbEntry->columnName, "start_time", sizeof( dbEntry->columnName ) );
  dbEntry->type          = int_4s;
  dbEntry->numberRows    = numEvents;
  dbEntry->rowDimensions = NULL;
  dbEntry->rows.int4s    = LALCalloc( dbEntry->numberRows,
      sizeof( *dbEntry->rows.int4s ) );
  if ( ! dbEntry->rows.int4s )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  thisEvent = eventList;
  for ( event = 0; event < numEvents; ++event )
  {
    dbEntry->rows.int4s[event] = thisEvent->startTimeNS / 1000000000;
    thisEvent = thisEvent->next;
  }


  /*
   *
   * Start time nanosecond residual.
   *
   */

  dbEntry->next           = LALCalloc( 1, sizeof( *dbEntry->next ) );
  if ( ! dbEntry->next )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  dbEntry->next->previous = dbEntry;
  dbEntry                 = dbEntry->next;

  strncpy( dbEntry->tableName, "sngl_ringdown", sizeof( dbEntry->tableName ) );
  strncpy( dbEntry->columnName, "start_time_ns",
      sizeof( dbEntry->columnName ) );
  dbEntry->type          = int_4s;
  dbEntry->numberRows    = numEvents;
  dbEntry->rowDimensions = NULL;
  dbEntry->rows.int4s    = LALCalloc( dbEntry->numberRows,
      sizeof( *dbEntry->rows.int4s ) );
  if ( ! dbEntry->rows.int4s )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  thisEvent = eventList;
  for ( event = 0; event < numEvents; ++event )
  {
    dbEntry->rows.int4s[event] = thisEvent->startTimeNS % 1000000000;
    thisEvent = thisEvent->next;
  }


  /*
   *
   * Duration.
   *
   */

  dbEntry->next           = LALCalloc( 1, sizeof( *dbEntry->next ) );
  if ( ! dbEntry->next )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  dbEntry->next->previous = dbEntry;
  dbEntry                 = dbEntry->next;

  strncpy( dbEntry->tableName, "sngl_ringdown", sizeof( dbEntry->tableName ) );
  strncpy( dbEntry->columnName, "duration", sizeof( dbEntry->columnName ) );
  dbEntry->type          = real_4;
  dbEntry->numberRows    = numEvents;
  dbEntry->rowDimensions = NULL;
  dbEntry->rows.real4    = LALCalloc( dbEntry->numberRows,
      sizeof( *dbEntry->rows.real4 ) );
  if ( ! dbEntry->rows.real4 )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  thisEvent = eventList;
  for ( event = 0; event < numEvents; ++event )
  {
    dbEntry->rows.real4[event] = thisEvent->duration;
    thisEvent = thisEvent->next;
  }


  /*
   *
   * Amplitude.
   *
   */

  dbEntry->next           = LALCalloc( 1, sizeof( *dbEntry->next ) );
  if ( ! dbEntry->next )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  dbEntry->next->previous = dbEntry;
  dbEntry                 = dbEntry->next;

  strncpy( dbEntry->tableName, "sngl_ringdown", sizeof( dbEntry->tableName ) );
  strncpy( dbEntry->columnName, "amplitude", sizeof( dbEntry->columnName ) );
  dbEntry->type          = real_4;
  dbEntry->numberRows    = numEvents;
  dbEntry->rowDimensions = NULL;
  dbEntry->rows.real4    = LALCalloc( dbEntry->numberRows,
      sizeof( *dbEntry->rows.real4 ) );
  if ( ! dbEntry->rows.real4 )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  thisEvent = eventList;
  for ( event = 0; event < numEvents; ++event )
  {
    dbEntry->rows.real4[event] = thisEvent->amplitude;
    thisEvent = thisEvent->next;
  }


  /*
   *
   * Frequency.
   *
   */

  dbEntry->next           = LALCalloc( 1, sizeof( *dbEntry->next ) );
  if ( ! dbEntry->next )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  dbEntry->next->previous = dbEntry;
  dbEntry                 = dbEntry->next;

  strncpy( dbEntry->tableName, "sngl_ringdown", sizeof( dbEntry->tableName ) );
  strncpy( dbEntry->columnName, "frequency", sizeof( dbEntry->columnName ) );
  dbEntry->type          = real_4;
  dbEntry->numberRows    = numEvents;
  dbEntry->rowDimensions = NULL;
  dbEntry->rows.real4    = LALCalloc( dbEntry->numberRows,
      sizeof( *dbEntry->rows.real4 ) );
  if ( ! dbEntry->rows.real4 )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  thisEvent = eventList;
  for ( event = 0; event < numEvents; ++event )
  {
    dbEntry->rows.real4[event] = thisEvent->frequency;
    thisEvent = thisEvent->next;
  }
  

  /*
   *
   * Quality.
   *
   */

  dbEntry->next           = LALCalloc( 1, sizeof( *dbEntry->next ) );
  if ( ! dbEntry->next )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  dbEntry->next->previous = dbEntry;
  dbEntry                 = dbEntry->next;

  strncpy( dbEntry->tableName, "sngl_ringdown", sizeof( dbEntry->tableName ) );
  strncpy( dbEntry->columnName, "q", sizeof( dbEntry->columnName ) );
  dbEntry->type          = real_4;
  dbEntry->numberRows    = numEvents;
  dbEntry->rowDimensions = NULL;
  dbEntry->rows.real4    = LALCalloc( dbEntry->numberRows,
      sizeof( *dbEntry->rows.real4 ) );
  if ( ! dbEntry->rows.real4 )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  thisEvent = eventList;
  for ( event = 0; event < numEvents; ++event )
  {
    dbEntry->rows.real4[event] = thisEvent->quality;
    thisEvent = thisEvent->next;
  }
  

  /*
   *
   * Mass.
   *
   */

  dbEntry->next           = LALCalloc( 1, sizeof( *dbEntry->next ) );
  if ( ! dbEntry->next )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  dbEntry->next->previous = dbEntry;
  dbEntry                 = dbEntry->next;

  strncpy( dbEntry->tableName, "sngl_ringdown", sizeof( dbEntry->tableName ) );
  strncpy( dbEntry->columnName, "mass", sizeof( dbEntry->columnName ) );
  dbEntry->type          = real_4;
  dbEntry->numberRows    = numEvents;
  dbEntry->rowDimensions = NULL;
  dbEntry->rows.real4    = LALCalloc( dbEntry->numberRows,
      sizeof( *dbEntry->rows.real4 ) );
  if ( ! dbEntry->rows.real4 )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  thisEvent = eventList;
  for ( event = 0; event < numEvents; ++event )
  {
    dbEntry->rows.real4[event] = thisEvent->mass;
    thisEvent = thisEvent->next;
  }


  /*
   *
   * SNR.
   *
   */

  dbEntry->next           = LALCalloc( 1, sizeof( *dbEntry->next ) );
  if ( ! dbEntry->next )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  dbEntry->next->previous = dbEntry;
  dbEntry                 = dbEntry->next;

  strncpy( dbEntry->tableName, "sngl_ringdown", sizeof( dbEntry->tableName ) );
  strncpy( dbEntry->columnName, "snr", sizeof( dbEntry->columnName ) );
  dbEntry->type          = real_4;
  dbEntry->numberRows    = numEvents;
  dbEntry->rowDimensions = NULL;
  dbEntry->rows.real4    = LALCalloc( dbEntry->numberRows,
      sizeof( *dbEntry->rows.real4 ) );
  if ( ! dbEntry->rows.real4 )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  thisEvent = eventList;
  for ( event = 0; event < numEvents; ++event )
  {
    dbEntry->rows.real4[event] = thisEvent->snr;
    thisEvent = thisEvent->next;
  }


  /*
   *
   * Confidence.
   *
   */

  dbEntry->next           = LALCalloc( 1, sizeof( *dbEntry->next ) );
  if ( ! dbEntry->next )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  dbEntry->next->previous = dbEntry;
  dbEntry                 = dbEntry->next;

  strncpy( dbEntry->tableName, "sngl_ringdown", sizeof( dbEntry->tableName ) );
  strncpy( dbEntry->columnName, "confidence", sizeof( dbEntry->columnName ) );
  dbEntry->type          = real_4;
  dbEntry->numberRows    = numEvents;
  dbEntry->rowDimensions = NULL;
  dbEntry->rows.real4    = LALCalloc( dbEntry->numberRows,
      sizeof( *dbEntry->rows.real4 ) );
  if ( ! dbEntry->rows.real4 )
  {
    ABORT( status, WRAPPERRINGSEARCHH_EALOC, WRAPPERRINGSEARCHH_MSGEALOC );
  }
  thisEvent = eventList;
  for ( event = 0; event < numEvents; ++event )
  {
    dbEntry->rows.real4[event] = thisEvent->confidence;
    thisEvent = thisEvent->next;
  }

  RETURN( status );
}
