/* -*- mode: c; c-basic-offset: 2; -*- */
/******** <lalVerbatim file="SlopeLALInitSearchC"> ********
Author: Daw, E
$Id: SlopeLALInitSearch.c,v 1.3 2003/01/21 23:55:39 edaw Exp $
********* </lalVerbatim> ********/

#ifdef DEBUG_SLOPE
#include <stdio.h>
#endif
#include <math.h>

/* functions for memory allocation */
#include <lal/LALStdlib.h>

/* functions for vector creation and manipulation */
/* #include <lal/AVFactories.h> */

/* functions for LAL MPI interface */
#include <lal/Comm.h>

/* functions for slope filtering and peak searching data */
#include <lal/SlopeDetectorFilter.h>

/* structures and definitions for this wrapper code */
#include <SlopeDatatypes.h>

NRCSID( SLOPELALINITSEARCHC, "$Id: SlopeLALInitSearch.c,v 1.3 2003/01/21 23:55:39 edaw Exp $");

/******** local static function declarations ********/

/****************************************************/
/** note that currently my code is designed to run **/
/** on a single machine. ****************************/
/****************************************************/

/******** wrapperAPI function definitions */

/* Initialize some parameters for the search and dynamically */
/* allocate some memory */

void LALInitSearch(LALStatus*           status,
		   void**               psearchParams,
		   LALInitSearchParams* initSearchParams)
{
  SlopeSearchParams* searchParams;
  /* these are the search input parameters */
  INT4 ac;
  CHAR** av;
  INT4 argcount;

  INITSTATUS( status, "SlopeLALInitSearch", SLOPELALINITSEARCHC );
  ATTATCHSTATUSPTR(status);

  /* get search input arguments */
  ac = initSearchParams->argc;
  av = initSearchParams->argv;

  /* allocate memory for search parameter buffer */
  searchParams = LALMalloc(sizeof(SlopeSearchParams));
  *psearchParams = (void*) searchParams;

  /* set search started bit to false */
  /* searchParams->searchStarted = 0; */

  /* initialize search parameters for simple search from defaults */

  strcpy(searchParams->channelName,
	 SLOPE_DEFAULT_CHANNEL);                    /*search channel*/

  strcpy(searchParams->ifoName,
	 SLOPE_DEFAULT_IFO);                        /*interferometer*/

  searchParams->samplingRate = 
    SLOPE_DEFAULT_SAMPLING_RATE ;                   /*sampling rate*/

  searchParams->forder = 
    SLOPE_DEFAULT_MEASURE_WIDTH ;                   /*bins for slope measure*/

  searchParams->peakWidth =                         /*number of adjacent*/
    SLOPE_DEFAULT_CORRELATED_WIDTH ;                /*filtered data bins*/
                                                    /*grouped under 1 peak*/
                                                    /*in peak search*/

  searchParams->triggerThreshold = 
    SLOPE_DEFAULT_THRESHOLD ;                       /*threshold for trigger*/

  searchParams->batchSize=BATCH_SIZE;
  searchParams->thresholdAdaptive = 0;


  /* go through the input paramters, overwriting some or all of */
  /* the above defaults with those present in the input parameters */

  for(argcount = 1; argcount < ac; ++argcount) {
    if (strcmp(av[argcount],"-c") == 0) {
      strcpy(searchParams->channelName,av[++argcount]);
    } else if (strcmp(av[argcount],"-smw") == 0) {
      sscanf(av[++argcount],"%u",
	     &(searchParams->forder));
    } else if (strcmp(av[argcount],"-ifo") == 0) {
      strcpy(searchParams->ifoName,av[++argcount]);
    } else if (strcmp(av[argcount],"-adapt") == 0) {
      if ( strcmp(av[++argcount],"yes") == 0 ) {
	searchParams->thresholdAdaptive = 1;
      }
    } else if (strcmp(av[argcount],"-pw") == 0) {
      sscanf(av[++argcount],"%u",
	     &(searchParams->peakWidth));
    } else if (strcmp(av[argcount],"-sr") == 0) {
      sscanf(av[++argcount],"%f",
	     &(searchParams->samplingRate));
    } else if (strcmp(av[argcount],"-t") == 0) {
      sscanf(av[++argcount],"%f",
	     &(searchParams->triggerThreshold));
    } else if (strcmp(av[argcount],"-sdtype") == 0) {
      sscanf(av[++argcount],"%u",
	     &(searchParams->function_select));
    } else if (strcmp(av[argcount],"-batchSize")==0){
      sscanf(av[++argcount],"%u",
	     &(searchParams->batchSize));
    } else {
      ABORT(status, INITSEARCH_BADINPUTARGS, INITSEARCH_BADINPUTARGSMESG);
    }
  }
  
#ifdef SLDEBUG
  printf("%s : %u : %u : %f : %f : %u\n",
	 searchParams->channelName,
	 searchParams->forder,
	 searchParams->peakWidth,
	 searchParams->samplingRate,
	 searchParams->triggerThreshold,
	 searchParams->function_select);
#endif

  /* calculate some important quantities from the input parameters to */
  /* the search */

  searchParams->sampling_period_s = 1 / searchParams->samplingRate;
  searchParams->waveform_offset = 0;
  
  searchParams->timeIntervalForSlopeMeasure =       /*interval for slope*/
    (REAL4)searchParams->forder /                   /*measure */
    searchParams->samplingRate ;

  searchParams->numberOfTriggersSoFar = 0;          /*counter for triggers*/

  /* allocate history buffer for filter input data */
  searchParams->nSampInFiltHistBuf = 
    searchParams->forder - 1;
  searchParams->filtHistBuf = 
    LALCalloc( searchParams->nSampInFiltHistBuf,
	       sizeof(REAL4) );
  searchParams->filtHistAllocated = 0;
  searchParams->history_allocated = LALMalloc( sizeof(UINT4) );
  *(searchParams->history_allocated) = searchParams->filtHistAllocated;
  searchParams->history = searchParams->filtHistBuf;

  /* if the filter is of a convolution type, allocate memory for the taps */
  if( ((searchParams->function_select == FILTER_OUTPUT_TAPS_SET_BOXCAR) ||
       (searchParams->function_select == FILTER_OUTPUT_TAPS_SET_GAUSSIAN)) ||
      (searchParams->function_select == FILTER_OUTPUT_TAPS_SET_SINE) ) {
    searchParams->taps_set = LALMalloc( sizeof(UINT4) );
    *(searchParams->taps_set) = 0;
    searchParams->tap = LALCalloc( searchParams->forder, sizeof(REAL4));
  }

  /* allocate history buffer for peak search output data */
  searchParams->nSampInPeakHistBuf =
    searchParams->peakWidth - 1;
  searchParams->peakHistBuf =
    LALCalloc( searchParams->nSampInPeakHistBuf,
	       sizeof(REAL4) );
  searchParams->peakHistAllocated = 0;

  /* initialize parameter for length of new data set */
  searchParams->nData = 0; 

  /* initialize event list */
  searchParams->firstEvent = 0x0;
  searchParams->currentEvent = 0x0;

  /* initialize mpi information */
  searchParams->mpiMaster = initSearchParams->nodeClass;

  searchParams->filteredData=NULL;
  searchParams->dataScanned=0;
  searchParams->dataNormalized=0;
  searchParams->summary=NULL;

  DETATCHSTATUSPTR(status);
  RETURN(status);
}









